
#include<stdio.h>
#include<stdlib.h>
#include<stddef.h>
#include<time.h>
#include<math.h>
#include<malloc.h>
#include<string.h>
#include"sslib.h"
// Global Parameters 
const int slots = 180;				//Nr of slots in histogram over wetting  angles after change. 0->1, 1->2, .... 179->180.
const int xLen = 40;
const int yLen = 40;
const int totalStepsA = 200000;
const int totalStepsB = 200000;
const int totalStepsC = 200000;
const int dStep = 100;		
const int saveBubbles = 0; 			// Will save the bubble distribution files (1) or not (0)
const int Samples = 5;				// Number of different samples
const int printout_wettabilityflips = 0;	// if ==1 : print out a pic of which links are flipped
const int printout_Flowdistr = 0;		// if ==1 : print out a pic of flow distr before/after wettabilitychange
const int printout_oilsat =0;			// if ==1 : print out a pic of oil/water saturation before/after wettabilitychange
const int Q_porevolumes= 150;	        	
const int coord = 4;
const int maxBubbles = 3;
const double saturationW = 0.1;			// oil sat. const in simulation. system to be initially oil wetting                               
const double thetacrit = 121.0;			// Value for Theta_min
const double linkLen = 1.0;			// mm
const double linkRadMin = 0.1;			// mm
const double linkRadMax = 0.4;			// mm
const double muWET = 0.01;			// 10 Pa*s = 100 Poise
const double muNON = 0.01;			// 10 Pa*s = 100 Poise
const double Ca = 1.0E-03;			// Capillary Number
const double surfaceTension = 3.0; 		// dyne/mm
const double globalPressureA = 20.0;		// dyne/mm^2 = 10 Pa
const double globalPressureB = 2000.0;		// dyne/mm^2 = 10 Pa
const double ERR = 1.0E-12;
const double TOL = 1.0E-12;

// Global Definitions
#define idum        4712
#define PI          (4*atan(1))
#define pi2          (8*atan(1))
#define M           (muNON/muWET)
#define totalNodes  (xLen*yLen)
#define totalLinks  (coord*xLen*yLen/2)
#define nStep       ((totalStepsA+totalStepsB+totalStepsC)/dStep)
#define tST         (totalStepsA/5)
#define critcosvalue (cos(thetacrit*pi2/360))

/* Function Prototypes */
void mem_Aloc();
void mem_Free();
void node_Connect_PB();
void initSys_Sequential();
void initSys_Sequential_gauss();
void initSys_RandomLink();
void initSys_RandomBubl();
void calc_qRate();
void calc_Link_Mob_Pc();
void solve_Conj_Grad(double pressure);
void calc_Flow();
void bubbles_Kill(int link);
void bubbles_Create(int link);
void bubbles_Merge_CM(int link);
void bubbles_Merge(int link);
void dynamic_Move();
void save_Parameters(char dir[80]);
void save_bblDist(char dirA[80]);
void draw_linkangles(char dir[80]);
void draw_LinkSat(char dir[80]);
void save_FlowDist(char dirA[80]);
void pic_Post(FILE *fp, int sx, int sy);
void ps_Header(FILE *fp, int xLen, int yLen, int sx, int sy);
double bubbles_Move(int link);
double random_LCG (double *seed);
void draw_Flowdistr(char dir[80], double qmax);		//with network lattice
void draw_Flowdistr2(char dir[80], double qmax);	//without network lattice
void countchangedlinks(char dir[80]);

// Global Variables
int timeStep;
int is; /// removerd form declaration in main() int is; 
char dir0[80], dir1[80], dir0F[80], dir2[80];

double seed = idum;
double volumeW, sTens, fluxTL, fluxW, flowTL, flowW, capA, capB, qRate; //qRate in mm^3/s.
double globalPressure, velMax, deltaT, totalT, flipT, dissipation, globalCond;
int *linkNodeUp, *linkNodeDn,  *numberOfBubbles, *linkIsGhost;
int **nodeNeighbour, **linkNeighbour;
double *linkRad, *nodePressure, *linkWFirst, *nodeFluxW, *nodeFluxTL, *mPos;
double *linkPc, *linkMob, *linkQ, *linkQ2, *linkAtil, *linkBtil, *CosTheta, *linkSatu, *q_beforechange, *q_afterchange, *q_final, *q_tot, *Q_thr,*w_angles, *histogram, *VolInj;
double *PresG, *PresC, *Dissp, *CondG, *FlowF, *TimeT, *Flow2;
double **bubbleStr, **bubbleEnd;



/*Main Program*/
int main() {
	

	int j, c, adr, nr, k, l, m, n, g, h, o, p, r, i, ip, link, flipNum, t1, t2, s, u, step;  
	double area, dTens, qmax, Qtot, porevolume;			//qmax used to create flow distr before/after wchange
	char fn1[80], fn2[80], fn3[80], fn4[80], fn5[80], fn6[80];
	double PresG_tm, PresC_tm, Dissp_tm, CondG_tm, FlowF_tm, TimeT_tm, Flow2_tm;
	FILE *fp1, *fp2, *fp3, *fp4, *fp5, *fp6, *fp8, *fp9, *fp10, *fp11;  

	mem_Aloc(); node_Connect_PB();

Qtot=0;
porevolume=0;	// used to calc nr of porevolumes injected. 
step=0;		





	sprintf(dir0,"%dx%d_NewAlg M-%.1e_Ca-%.2e_S-%.3f", xLen, yLen, M, Ca, saturationW);
	if (totalStepsB > 0)
		sprintf(dir0F,"%s/Thcr-%.1f", dir0, thetacrit);
	else
		sprintf(dir0F,"%s/steady", dir0);
	sprintf(dir1,"%s/bubbleDst",dir0F);
	mkdir(dir0,0755); mkdir(dir0F,0755); if (saveBubbles) mkdir(dir1,0755);
	sprintf(fn1,"%s/pressure.d",dir0F); fp1 = fopen(fn1,"w");
	sprintf(fn2,"%s/fracFlow.d",dir0F); fp2 = fopen(fn2,"w");
	sprintf(fn3,"%s/conductG.d",dir0F); fp3 = fopen(fn3,"w");
	sprintf(fn4,"%s/d issiptn.d",dir0F); fp4 = fopen(fn4,"w");
	sprintf(fn5,"%s/realTime.d",dir0F); fp5 = fopen(fn5,"w");
	sprintf(fn6,"%s/dirnFlow.d",dir0F); fp6 = fopen(fn6,"w");
	

	is = 0;
	
	samp:
	
	step=0;  //reset step before each sample.
	Qtot=0;
	porevolume=0;
	i = 0; timeStep = 0; totalT = 0.0;
	sTens = 0.0;
	dTens = surfaceTension/tST;
	t1 = 1; t2 = tST; 
	initSys_Sequential();
	//initSys_Sequential_gauss();  //Gaussian radii distr

for(u=0; u<totalLinks; u++){
Q_thr[u]=Q_porevolumes*PI*linkRad[u] * linkRad[u] * linkLen;   // Calc Q_thr after system is initialized
porevolume=porevolume + PI*linkRad[u] * linkRad[u] * linkLen;	// total pore volume of network
}                                       				
	calc_qRate();						
	if (saveBubbles && is==0) save_bblDist(dir1);

	tick:
	PresG_tm = 0.0; PresC_tm = 0.0; FlowF_tm = 0.0;
	CondG_tm = 0.0; Dissp_tm = 0.0; TimeT_tm = 0.0; Flow2_tm = 0.0;
	for (timeStep=t1; timeStep<=t2; timeStep++) {


Qtot=Qtot + qRate*deltaT;					// total volume injected


if(timeStep%dStep==0){
VolInj[step]=VolInj[step]+Qtot/porevolume;
step++;
}



if(timeStep>(totalStepsA-10000) && timeStep<totalStepsA){		//To sum up flowdistr in each tube before wettability change 
for(h=0; h<totalLinks; h++){
q_beforechange[h] = q_beforechange[h]+fabs(linkQ[h]*deltaT);
}
}



if(timeStep>(totalStepsA+totalStepsB) && timeStep<(totalStepsA+totalStepsB+10000)){		//To sum up flowdistr after wettability change
for(o=0; o<totalLinks; o++){
q_afterchange[o] = q_afterchange[o]+fabs(linkQ[o]*deltaT);
}
}

if(timeStep>(totalStepsA+totalStepsB+totalStepsC -10000) && timeStep<(totalStepsA+totalStepsB+totalStepsC)){	//To sum up flowdistr at 
for(r=0; r<totalLinks; r++){											//final steady state
q_final[r] = q_final[r]+fabs(linkQ[r]*deltaT);
}
}



if(timeStep==totalStepsA && printout_oilsat ==1) draw_LinkSat(dir0F);	                     //Draw Link sat picture before Wettability alterations
	





if(timeStep>totalStepsA && timeStep<totalStepsA+totalStepsB){       //max flowrate to be used for logscale plot of flowdistr
qmax=0;
for(s=0; s<totalLinks; s++){
if(q_beforechange[s]>qmax){
qmax=q_beforechange[s];
}
}
}


						      /// Calc the wetting angle in at all timesteps between timestep A and C, for t>C, keep same value
if(timeStep>totalStepsA && timeStep<(totalStepsA+totalStepsB)){
for (k=0; k<totalLinks; k++){
		q_tot[k] = q_tot[k]+fabs(linkQ[k])*(1-linkSatu[k])*(deltaT);	/// to only add up the amount of water flowing in the tubes.//	
		CosTheta[k] = (2/PI)*atan((20/Q_thr[k])*(q_tot[k] - Q_thr[k])); /// linkSatu[i] is oil saturation in link i. use oilsat+wsat =1
		if(CosTheta[k]>critcosvalue){CosTheta[k]=critcosvalue;}		//defines minimum wetting angle, theta_min.   
									
		}
	}	



if(printout_wettabilityflips==1){
	if(timeStep>totalStepsA && timeStep<(totalStepsA+totalStepsB)){		// print wetting angles in each tube every 10.000 timesteps 
	if(timeStep % 10000 == 0){draw_linkangles(dir0F);} 			// during wettability altering part of simulation
	}
}


		sTens += dTens;
		calc_Flow();
		dynamic_Move();
		PresG_tm += globalPressure;
		PresC_tm += -capB/capA;
		FlowF_tm += fluxW/fluxTL;
		CondG_tm += globalCond;
		Dissp_tm += dissipation;
		TimeT_tm += totalT;
		Flow2_tm += flowW/flowTL;
		

		
		if (timeStep%dStep == 0) {
			if (saveBubbles && is==0) save_bblDist(dir1);
			PresG[i] += PresG_tm/dStep;
			PresC[i] += PresC_tm/dStep;
			FlowF[i] += FlowF_tm/dStep;
			CondG[i] += CondG_tm/dStep;
			Dissp[i] += Dissp_tm/dStep;
			TimeT[i] += TimeT_tm/dStep;
			Flow2[i] += Flow2_tm/dStep;
			i++;
			PresG_tm = 0.0; PresC_tm = 0.0; FlowF_tm = 0.0;
			CondG_tm = 0.0; Dissp_tm = 0.0; TimeT_tm = 0.0; Flow2_tm = 0.0;
		}
	}


	
	if (t2 == tST) {
		t1 = t2 + 1; t2 = totalStepsA;
		dTens = 0.0; sTens = surfaceTension;
		goto tick;
	}

 

	else if (totalStepsB > 0 && t2 == totalStepsA ) {  


		t1 = t2 +1; t2 = totalStepsA+totalStepsB;		
		flipT = totalT;
		goto tick;
	}





	else if (totalStepsC > 0 && t2 == (totalStepsA+totalStepsB)) {		

		t1 = t2 + 1; t2 = totalStepsA + totalStepsB + totalStepsC;

		goto tick;
	}



//// Program finished first sample

//printout of angle distribution after wettability alteration:

for(i=0;i<totalLinks;i++){
w_angles[i]=(acos(CosTheta[i]))*360/(2*PI);
}
	
if(printout_Flowdistr == 1){		///////////Print flowdistr.
draw_Flowdistr(dir0F, qmax);		// with network lattice
//draw_Flowdistr2(dir0F, qmax);		//without network lattice
}	

	is ++;
	if (is < Samples){

	for(p=0; p<totalLinks; p++){
	q_beforechange[p]=0;
	q_afterchange[p]=0;
	q_final[p]=0;
	q_tot[p]=0;									
	}
 goto samp;
}




if(printout_oilsat ==1)draw_LinkSat(dir0F); //print picture of oil saturation in each link. black= 100% oil
//draw_LinkSat(dir0F);


fp9 = fopen("angledistribution.d","a+");		
	for(i=0; i<totalLinks; i++){
		
		for(j=1;j<slots;j++){
			if (w_angles[i]<j){
			histogram[j]=histogram[j] +1;
			goto test;
			}

		}
		
		test:
		
		fprintf (fp9, "%f\t%d\n", w_angles[i], i);

		}
fclose(fp9);

	
fp10 = fopen("histogram2.d","a+");
	for(i=0; i<slots; i++){
		fprintf (fp10, "%d\t%f\n", i, histogram[i]);
		}
fclose(fp10);



	for (i=0; i<nStep; i++) {
		ip = (i+1)*dStep;
		fprintf(fp1,"%f\t%e\t%e\n", VolInj[i]/Samples, PresG[i]/Samples, PresC[i]/Samples);
		fprintf(fp2,"%f\t%e\n"    , VolInj[i]/Samples, FlowF[i]/Samples);		
		fprintf(fp3,"%d\t%e\n"    , ip, CondG[i]/Samples);
		fprintf(fp4,"%d\t%e\n"    , ip, Dissp[i]/Samples);
		fprintf(fp5,"%d\t%e\n"    , ip, TimeT[i]/Samples);
		fprintf(fp6,"%d\t%e\n"    , ip, Flow2[i]/Samples);
	}
	
	save_Parameters(dir0F);
	fclose(fp1); fclose(fp2); fclose(fp3); fclose(fp4); fclose(fp5);
	mem_Free();
}



/*Allocating memory, making arrays*/
void mem_Aloc() {
	nodeNeighbour = initInt2D(totalNodes, coord, 0);
	linkNeighbour = initInt2D(totalNodes, coord, 0);
	linkNodeUp = initInt1D(totalLinks, 0);
	linkNodeDn = initInt1D(totalLinks, 0);
	linkIsGhost = initInt1D(totalLinks, 0);
	linkRad = initReal1D(totalLinks, 0.0);
	bubbleStr = initReal2D(totalLinks, maxBubbles+2, 0.0);
	bubbleEnd = initReal2D(totalLinks, maxBubbles+2, 0.0);
	numberOfBubbles = initInt1D(totalLinks, 0);
	nodePressure = initReal1D(totalNodes, 0.0);
	linkQ = initReal1D(totalLinks, 0.0);
	linkQ2 = initReal1D(totalLinks, 0.0);
	linkPc = initReal1D(totalLinks, 0.0);
	linkMob = initReal1D(totalLinks, 0.0);
	linkSatu = initReal1D(totalLinks, 0.0);
	linkAtil = initReal1D(totalLinks, 0.0);
	linkBtil = initReal1D(totalLinks, 0.0);
	linkWFirst = initReal1D(totalLinks, 0.0); 
	nodeFluxW = initReal1D(totalNodes, 0.0);
	nodeFluxTL = initReal1D(totalNodes, 0.0);
	CosTheta = initReal1D(totalLinks, 1.0);
	PresG = initReal1D(nStep, 0.0);
	PresC = initReal1D(nStep, 0.0);
	Dissp = initReal1D(nStep, 0.0);
	CondG = initReal1D(nStep, 0.0);
	FlowF = initReal1D(nStep, 0.0);
	Flow2 = initReal1D(nStep, 0.0);
	TimeT = initReal1D(nStep, 0.0);
	mPos = initReal1D((maxBubbles+4)*2, 0.0);
	q_beforechange = initReal1D(totalLinks, 0);                    
	q_afterchange = initReal1D(totalLinks, 0);   
	q_final = initReal1D(totalLinks, 0);                 
	q_tot = initReal1D(totalLinks, 0);
	Q_thr = initReal1D(totalLinks, 0);                   
	w_angles = initReal1D(totalLinks,0);			
	histogram = initReal1D(slots,0);	
	VolInj=initReal1D(nStep,0.0);			
}

/*Freeing memory, at the end of program*/
void mem_Free() {
	free(linkNodeUp);
	free(linkNodeDn);
	free(linkIsGhost);
	free(linkRad);
	free(numberOfBubbles);
	free(nodePressure);
	free(linkQ);
	free(linkQ2);
	free(linkPc);
	free(linkMob);
	free(linkAtil);
	free(linkBtil);
	free(linkSatu);
	free(linkWFirst); 
	free(nodeFluxW);
	free(nodeFluxTL);
	free(CosTheta);
	free(mPos);
	free(PresG);
	free(PresC);
	free(Dissp);
	free(CondG);
	free(FlowF);
	free(Flow2);
	free(TimeT);
	freeInt2D(nodeNeighbour, totalNodes);
	freeInt2D(linkNeighbour, totalNodes);
	freeReal2D(bubbleStr, totalLinks);
	freeReal2D(bubbleEnd, totalLinks);
}

/*Making the Network, defining nearest neighbour nodes and links of each node and link*/
void node_Connect_PB() {
	int i,j,k;
	//Assigning the Node neighbours of a Node
	for (j=0; j<yLen; j+=2) {
		for (i=0; i<xLen; i++) {
			k = j*xLen + i;
			nodeNeighbour[k][0] = k - xLen - 1;
			nodeNeighbour[k][1] = k - xLen;
			nodeNeighbour[k][2] = k + xLen;
			nodeNeighbour[k][3] = k + xLen - 1;
			if (k < xLen) {
				nodeNeighbour[k][0] = nodeNeighbour[k][0] + totalNodes;
				nodeNeighbour[k][1] = nodeNeighbour[k][1] + totalNodes;
			}
			if (i == 0) {
				nodeNeighbour[k][0] = nodeNeighbour[k][0] + xLen;
				nodeNeighbour[k][3] = nodeNeighbour[k][3] + xLen;
			}
		}
	}

	for (j=1; j<yLen; j+=2) {
		for (i=0; i<xLen; i++) {
			k = j*xLen + i;
			nodeNeighbour[k][0] = k - xLen;
			nodeNeighbour[k][1] = k - xLen + 1;
			nodeNeighbour[k][2] = k + xLen + 1;
			nodeNeighbour[k][3] = k + xLen;
			if (i == (xLen-1)) {
				nodeNeighbour[k][1] = nodeNeighbour[k][1] - xLen;
				nodeNeighbour[k][2] = nodeNeighbour[k][2] - xLen;
			}
			if (k >= (yLen-1)*xLen) {
				nodeNeighbour[k][2] = nodeNeighbour[k][2] - totalNodes;
				nodeNeighbour[k][3] = nodeNeighbour[k][3] - totalNodes;
			}
		}
	}

	//Assigning the Link neighbours of a Node

	//Direction 0, 1 (down-left, down-right)
	for (k=0; k<totalNodes; k++) {
		linkNeighbour[k][0] = 2*k;
		linkNeighbour[k][1] = 2*k + 1;
	}
	//Direction 2, 3 (upper-right, upper-left)
	for (k=0; k<totalNodes; k++) {
		linkNeighbour[k][2] = linkNeighbour[nodeNeighbour[k][2]][0];
		linkNeighbour[k][3] = linkNeighbour[nodeNeighbour[k][3]][1];
		linkNodeUp[linkNeighbour[k][0]] = k;
		linkNodeDn[linkNeighbour[k][0]] = nodeNeighbour[k][0];
		linkNodeUp[linkNeighbour[k][1]] = k;
		linkNodeDn[linkNeighbour[k][1]] = nodeNeighbour[k][1];
	}
	for (k=0; k<(2*xLen); k++)
		linkIsGhost[k] = 1;
}

/*Random Number generator*/
double random_LCG (double *seed) {
	double b32,random;
	b32=2147483647;
	*seed = *seed*16807;
	//*seed = *seed*time(NULL);
	*seed = fmod(*seed,b32);
	random = *seed/b32;
	return random;
}

/*Initializing system, Two bands of the two fluids*/
void initSys_Sequential() {
	int i;
	double random, volTL, volW, fracW;
	for (i=0; i<totalNodes; i++) {
		random = random_LCG(&seed);
		nodePressure[i] = random;
	}
	volTL = 0.0;
	for (i=0; i<totalLinks; i++) {
		random = random_LCG(&seed);
		linkRad[i] = linkLen * (linkRadMin + (linkRadMax - linkRadMin) * random);
		linkWFirst[i] = 0.0;
		numberOfBubbles[i] = 0;
		CosTheta[i] = -0.95;    //acos(165), starting point at wetting angle theta = 165 degrees
		volTL += linkRad[i] * linkRad[i] * linkLen;
	}
	volW = 0.0; fracW = 0.0; i = 0;
	while (fracW<saturationW) {
		bubbleStr[i][0] = 0.0;
		bubbleEnd[i][0] = linkLen;
		numberOfBubbles[i] = 1;
		linkSatu[i] = 1.0;
		volW += linkRad[i] * linkRad[i] * linkLen;
		fracW = volW/volTL;
		i++;
	}
	if (fracW>saturationW) {
		i--;
		fracW = (volW - saturationW*volTL)/(linkRad[i] * linkRad[i]);
		bubbleEnd[i][0] = linkLen - fracW;
		linkSatu[i] = bubbleEnd[i][0]/linkLen;
	}
}


void initSys_Sequential_gauss() {
	int i,adress;
	double random, volTL, volW, fracW,rand1,rand2,t1,t2,randlist[totalLinks];
	for (i=0; i<totalNodes; i++) {
		random = random_LCG(&seed);

		nodePressure[i] = random;
	}
	volTL = 0.0;



for(i=0;i<totalLinks/2;i++){
	rand1=random_LCG(&seed);	
	rand2=random_LCG(&seed);
	t1=sqrt(-pi2*log(rand1))*cos(pi2*rand2)*0.1 +0.5;
	t2=sqrt(-pi2*log(rand1))*sin(pi2*rand2)*0.1 +0.5;	
	adress=2*i;
	randlist[adress]=t1;
	randlist[adress+1]=t2;
	}

	for (i=0; i<totalLinks; i++) {
		if(randlist[i]>0.1){
		linkRad[i] = linkLen * (linkRadMin + (linkRadMax - linkRadMin) * randlist[i]);
		}   						//if rad<0.1, set to 0.1
		else{linkRad[i]=0.1;}
		linkWFirst[i] = 0.0;
		numberOfBubbles[i] = 0;
		CosTheta[i] = -1.0;
		volTL += linkRad[i] * linkRad[i] * linkLen;
	}
	volW = 0.0; fracW = 0.0; i = 0;
	while (fracW<saturationW) {
		bubbleStr[i][0] = 0.0;
		bubbleEnd[i][0] = linkLen;
		numberOfBubbles[i] = 1;
		linkSatu[i] = 1.0;
		volW += linkRad[i] * linkRad[i] * linkLen;
		fracW = volW/volTL;
		i++;
	}
	if (fracW>saturationW) {
		i--;
		fracW = (volW - saturationW*volTL)/(linkRad[i] * linkRad[i]);
		bubbleEnd[i][0] = linkLen - fracW;
		linkSatu[i] = bubbleEnd[i][0]/linkLen;
	}
}


/*Initializing system, Randomly selected completely filled links*/
void initSys_RandomLink() {
      int i;
      double random, volTL, volW, fracW;
      for (i=0; i<totalNodes; i++) {
	     random = random_LCG(&seed);
	     nodePressure[i] = random;
      }
      volTL = 0.0;
      for (i=0; i<totalLinks; i++) {
	     random = random_LCG(&seed);
	     linkRad[i] = linkLen * (linkRadMin + (linkRadMax - linkRadMin) * random);
	     linkWFirst[i] = 0.0;
	     numberOfBubbles[i] = 0;
	     CosTheta[i] = -1.0;
	     volTL += linkRad[i] * linkRad[i] * linkLen;
      }
      volW = 0.0; fracW = 0.0;          
      while (fracW<saturationW) {
	     i = totalLinks * random_LCG(&seed);
	     if (numberOfBubbles[i] == 0) { 
		    bubbleStr[i][0] = 0.0;
		    bubbleEnd[i][0] = linkLen;
		    numberOfBubbles[i] = 1;
		    linkSatu[i] = 1.0;
		    volW += linkRad[i] * linkRad[i] * linkLen;
		    fracW = volW/volTL;
	     }
      }
      if (fracW>saturationW) {
	     fracW = (volW - saturationW*volTL)/(linkRad[i] * linkRad[i]);
	     bubbleEnd[i][0] = linkLen - fracW;
	     linkSatu[i] = bubbleEnd[i][0]/linkLen;
      }
}

/*Initializing system, random bubbles in each link*/
void initSys_RandomBubl() {
      int i;
      double random;
      for (i=0; i<totalNodes; i++) {
	     random = random_LCG(&seed);
	     nodePressure[i] = random;
      }
      for (i=0; i<totalLinks; i++) {
	     random = random_LCG(&seed);
	     linkRad[i] = linkLen * (linkRadMin + (linkRadMax - linkRadMin) * random);
	     linkWFirst[i] = 0.0;
	     do {
	            random = random_LCG(&seed);
		    bubbleStr[i][0] = random*linkLen*(1.0-saturationW);
		    bubbleEnd[i][0] = bubbleStr[i][0] + (linkLen*saturationW);
	     } while (bubbleEnd[i][0] > linkLen);
	     numberOfBubbles[i] = 1;
	     CosTheta[i] = -1.0;
      }
}

/*Calculation total flow rate from given capillary number*/
void calc_qRate() {
	int i;
	double area;
	area = 0.0;
	for (i=0; i<totalLinks; i++)
		area += PI * linkRad[i] * linkRad[i] * linkLen;
	area = area/yLen;
	if (muNON > muWET) qRate = Ca * (surfaceTension * area) / muNON;
	else qRate = Ca * (surfaceTension * area) / muWET;
}

/*Calculation of link mobility and capillary pressure in a link*/              
void calc_Link_Mob_Pc() {
	int i, j;
	double temp, str, end, invLen, invRad, muEff;
	invLen = 1.0/linkLen;
	for (i=0; i<totalLinks; i++) {
		linkPc[i] = 0.0;
		invRad = CosTheta[i] / linkRad[i];			
		temp=0.0;
		for (j=0; j<numberOfBubbles[i]; j++) {
			str = bubbleStr[i][j];
			end = bubbleEnd[i][j];
			temp += end - str;
			linkPc[i] += 2.0 * sTens * invRad * (- cos(2.0 * PI * invLen * end) + cos(2.0 * PI * invLen * str));
		}
		temp = temp/linkLen;
		muEff = muNON * temp + muWET * (1.0 - temp);
		linkMob[i] = PI * pow(linkRad[i], 4) / (8.0 * muEff * linkLen);
	}
}

/*Conjugate gradient solver*/
void solve_Conj_Grad(double pressure) {
	int i, j, linkNb, nodeNb, counter;
	double am, bm, rps, rpn, pi, ri, api, ghost;
	double *p, *r, *ap;

	p = malloc(totalNodes * sizeof(double));
	r = malloc(totalNodes * sizeof(double));
	ap = malloc(totalNodes * sizeof(double));

	rpn = 0.0;
	//Initialization (Eq 32)
	for (i=0; i<totalNodes; i++) {
		p[i] = 0.0;
		for (j=0; j<coord; j++) {
			linkNb = linkNeighbour[i][j];
			nodeNb = nodeNeighbour[i][j];
			ghost = linkIsGhost[linkNb];
			if (nodeNb < i)
				p[i] -= linkMob[linkNb] * (nodePressure[nodeNb] - nodePressure[i] - (1.-2.*ghost)*linkPc[linkNb] - pressure*ghost);
			else if (nodeNb > i)
				p[i] -= linkMob[linkNb] * (nodePressure[nodeNb] - nodePressure[i] + (1.-2.*ghost)*linkPc[linkNb] + pressure*ghost);
		}
		r[i] = p[i];
		ri = r[i];
		rpn += ri*ri;
	}

	//Detemining whether we have the solution or not - Eq. 38 in BH.
	while(rpn >= ERR) {
		rps = rpn;
		//Eq. 33
		am = 0.0;
		for (i=0; i<totalNodes; i++) {
			api = 0.0;
			pi = p[i];
			for (j=0; j<coord; j++) {
				linkNb = linkNeighbour[i][j];
				nodeNb = nodeNeighbour[i][j];
				api += linkMob[linkNb] * (p[nodeNb] - pi);
			}
			ap[i] = api;
			am += pi * api;
		}
		am = rps / am;
		rpn = 0.0;
		for (i=0; i<totalNodes; i++) {
			nodePressure[i] += am * p[i]; 	//Eq. 34
			r[i] -= am * ap[i];			//Eq. 35
			ri = r[i];
			rpn += ri * ri;
		}
		bm = rpn / rps;					//Eq. 36
		for (i=0; i<totalNodes; i++) {
			p[i] = r[i] + bm*p[i];			//Eq. 37
		}
	}
	free(p);
	free(r);
	free(ap);
}

/*Calculating local flow rates by linear extrapolation, Q = AP + B*/
 void  calc_Flow() {
	int i, j, k;
	double qRate1, qRate2, dP;
	dP = globalPressureA - globalPressureB;
	calc_Link_Mob_Pc();
	

	solve_Conj_Grad(globalPressureA);
	qRate1 = 0.0;
	for (i=0; i<totalLinks; i++) {
		linkQ[i] = -linkMob[i]*(nodePressure[linkNodeUp[i]] - nodePressure[linkNodeDn[i]] + linkPc[i] - globalPressureA * linkIsGhost[i]);
		qRate1 = qRate1 + linkQ[i] * linkIsGhost[i];
	}
	
	solve_Conj_Grad(globalPressureB);
	qRate2 = 0.0;
	for (i=0; i<totalLinks; i++) {
		linkQ2[i] = -linkMob[i]*(nodePressure[linkNodeUp[i]] - nodePressure[linkNodeDn[i]] + linkPc[i] - globalPressureB * linkIsGhost[i]);
		qRate2 = qRate2 + linkQ2[i] * linkIsGhost[i];
		linkAtil[i] = (linkQ[i] - linkQ2[i]) / dP;
	}
	
	capA = (qRate1 - qRate2) / dP;
	capB = qRate1 - capA * globalPressureA;
	globalPressure = qRate / capA - capB / capA;
	globalCond = qRate / globalPressure;

	dissipation = 0.0;
	for (i=0; i<totalLinks; i++) {
		linkBtil[i] = linkQ[i] - linkAtil[i] * globalPressureA;
		linkQ[i] = linkAtil[i]*globalPressure + linkBtil[i];
		dissipation += linkQ[i] * linkQ[i] / linkMob[i];
			
	}

	//dissipation = dissipation * 1.0E-08; //Conversion to SI unit (watt / J.s-1 / kg.mt2.s-3)
}

/*Moving bubbles according to local flow rate*/
double bubbles_Move(int link) {
	int i;
	double x, area, volW, lenW;
	area = PI * linkRad[link] * linkRad[link];
	volW = 0.0;
	linkSatu[link] = 0.0;
	if (linkQ[link] >= 0.0) {
		for(i=0; i<numberOfBubbles[link]; i++) {
			x = bubbleEnd[link][i] - bubbleStr[link][i];
			linkSatu[link] += x/linkLen;
			volW += x*area;
			bubbleStr[link][i] += linkQ[link] * deltaT / area;
			bubbleEnd[link][i] += linkQ[link] * deltaT / area;
			lenW = 0.0;
			if (bubbleStr[link][i]>linkLen && bubbleEnd[link][i]>linkLen) lenW = bubbleEnd[link][i] - bubbleStr[link][i];
			else if (bubbleStr[link][i]<=linkLen && bubbleEnd[link][i]>linkLen) lenW = bubbleEnd[link][i] - linkLen;
			nodeFluxW[linkNodeUp[link]] += lenW * area;
		}
		nodeFluxTL[linkNodeUp[link]] += linkQ[link] * deltaT;
	}
	else if (linkQ[link] < 0.0) {
		for (i=0; i<numberOfBubbles[link]; i++) {
			x = bubbleEnd[link][i] - bubbleStr[link][i];
			linkSatu[link] += x/linkLen;
			volW += x*area;
			bubbleStr[link][i] += linkQ[link] * deltaT / area;
			bubbleEnd[link][i] += linkQ[link] * deltaT / area;
			lenW = 0.0;
			if (bubbleStr[link][i]<0.0 && bubbleEnd[link][i]<0.0) lenW = bubbleEnd[link][i] - bubbleStr[link][i];
			else if (bubbleStr[link][i]<0.0 && bubbleEnd[link][i]>=0.0) lenW = -bubbleStr[link][i];
			nodeFluxW[linkNodeDn[link]] += lenW * area;
			flowW += linkQ[i]*(bubbleEnd[link][i] - bubbleStr[link][i]);
		}
		nodeFluxTL[linkNodeDn[link]] -= linkQ[link] * deltaT;
	}
	return volW;
}

/*Kill bubble, if too close*/
void bubbles_Kill(int link) {
	int i, j, bubbles;
	
	for (i=0; i<numberOfBubbles[link]; i++) {
		if (bubbleStr[link][i] < 0.0) bubbleStr[link][i] = 0.0;
		if (bubbleStr[link][i] > linkLen) bubbleStr[link][i] = linkLen;
		if (bubbleEnd[link][i] < 0.0) bubbleEnd[link][i] = 0.0;
		if (bubbleEnd[link][i] > linkLen) bubbleEnd[link][i] = linkLen;
	}

	bubbles = numberOfBubbles[link];
	for (i=numberOfBubbles[link]-1; i>=0; i--) {
		if (fabs(bubbleEnd[link][i]-bubbleStr[link][i]) < TOL) {
			bubbles--;
			for (j=i; j<bubbles; j++) {
				bubbleStr[link][j] = bubbleStr[link][j+1];
				bubbleEnd[link][j] = bubbleEnd[link][j+1];
			}
		}
	}
	numberOfBubbles[link] = bubbles;
}

/*Create new bubble when fluid enters new link*/
void bubbles_Create(int link) {
	int i;
	double str, end, fracDn, fracUp, area, dif;
	area = PI * linkRad[link] * linkRad[link];
	
	fracDn = nodeFluxW[linkNodeDn[link]] / nodeFluxTL[linkNodeDn[link]];
	if (fracDn > 1.0) fracDn = 1.0;
	fracUp = nodeFluxW[linkNodeUp[link]] / nodeFluxTL[linkNodeUp[link]];
	if (fracUp > 1.0) fracUp = 1.0;
	
	if (fabs(linkQ[link]) > TOL) {
		if (linkQ[link] >= 0.0) {
			str = linkWFirst[link] * linkQ[link] * deltaT * (1.0-fracDn) / area;
			end = str + linkQ[link] * deltaT * fracDn / area;
			for (i=numberOfBubbles[link]-1; i>=0; i--)  {
				bubbleStr[link][i+1] = bubbleStr[link][i];
				bubbleEnd[link][i+1] = bubbleEnd[link][i];
			}
			bubbleStr[link][0] = str;
			bubbleEnd[link][0] = end;
		}
		else if (linkQ[link] < 0.0) {
			dif = linkQ[link] * deltaT * (1.0-fracUp) / area;
			str = linkLen + linkWFirst[link] * dif + linkQ[link] * deltaT * fracUp / area;
			end = linkLen + linkWFirst[link] * dif;
			bubbleStr[link][numberOfBubbles[link]] = str;
			bubbleEnd[link][numberOfBubbles[link]] = end;
		}
		linkWFirst[link] = 1.0 - linkWFirst[link];
		numberOfBubbles[link] ++;
	}
}

/*Merge two bubbles if number to bubbles in a link exceeds the maximum given number*/
/*Old Rule*/
void bubbles_Merge(int link) {
	int i, bubble1, bubble2;
	double mindist;
	mindist = linkLen + TOL;
	bubble1 = -1;
	bubble2 = -1;
	
	for (i=1; i<numberOfBubbles[link]; i++) {
		if (bubbleStr[link][i] - bubbleEnd[link][i-1] < mindist) {
			mindist = bubbleStr[link][i] - bubbleEnd[link][i-1];
			bubble1 = i-1;
			bubble2 = i;
		}
	}
	if (bubble1 > -1 && bubble2 > -1) {
		if (linkQ[link] >= 0.0) {
			bubbleEnd[link][bubble1] = bubbleEnd[link][bubble2] - mindist;
			for (i=bubble2; i<numberOfBubbles[link]-1; i++) {
				bubbleStr[link][i] = bubbleStr[link][i+1];
				bubbleEnd[link][i] = bubbleEnd[link][i+1];
			}
		}
		else if (linkQ[link] < 0.0) {
			bubbleStr[link][bubble2] = bubbleStr[link][bubble1] + mindist;
			for (i=bubble1; i<numberOfBubbles[link]-1; i++) {
				bubbleStr[link][i] = bubbleStr[link][i+1];
				bubbleEnd[link][i] = bubbleEnd[link][i+1];
			}
		}
		numberOfBubbles[link] --;
	}
}

/*Merge two bubbles if number to bubbles in a link exceeds the maximum given number*/
/*New Rule, keeping CM fixed*/
void bubbles_Merge_CM(int link) {
	int i, j, k, iob, inb, imn, delPos;
	double x, dL, r1, r2, r12, m1, m2;
	iob = numberOfBubbles[link];
	imn = -1;
	delPos = -1;
	dL = linkLen + TOL;
	mPos[++imn] = 0.0;
	mPos[++imn] = 0.0;
	for (i=0; i<iob; i++) {
		mPos[++imn] = bubbleStr[link][i];
		mPos[++imn] = bubbleEnd[link][i];
		x = mPos[imn] - mPos[imn-1];
		if (dL>x && mPos[imn-1]>TOL && (linkLen-mPos[imn])>TOL) {
			dL = x;
			delPos = imn - 1;
		}
	}
	mPos[++imn] = linkLen;
	mPos[++imn] = linkLen;
	for (i=3;i<imn-3;i+=2) {
		x = mPos[i+1] - mPos[i];
		if (dL>x) {
			dL = x;
			delPos = i;
		}
	}
	if (delPos>-1) {
		r1 = mPos[delPos] - mPos[delPos-1];
		r2 = mPos[delPos+2] - mPos[delPos+1];
		r12 = r1+r2;
		m1 = dL*r2 / r12;
		m2 = dL*r1 / r12;
		mPos[delPos-1] += m1;
		mPos[delPos+2] -= m2;
	}
	for (i=delPos; i<imn-1; i++) mPos[i] = mPos[i+2];
	inb = 0;
	j=2; k=imn-4;
	if (delPos == 2) j = 0;
	else if (delPos == imn-3) k = imn-2;
	for (i=j;i<k;i+=2) {
		bubbleStr[link][inb] = mPos[i];
		bubbleEnd[link][inb] = mPos[i+1];
		inb++;
	}
	numberOfBubbles[link] = inb;
}

/*One complete Bubble dynamics, collection of 4 bubble rules, bubble Move, Kill, Create and Merge*/
void dynamic_Move() {
	int i, velMaxPos;
	double vel, area, temp;

	velMax = 0.0;
	velMaxPos = -1;
	for (i=0; i<totalLinks; i++) {
		area = PI * linkRad[i] * linkRad[i];
		vel = linkQ[i] / area;
		if (velMax < fabs(vel) && numberOfBubbles[i] > 0) {
			velMax = fabs(vel);
			velMaxPos = i;

		}
	}
	if (velMaxPos == -1) {
		printf("No maximum velocity found.\n");
		exit(1);
	}
	deltaT = 0.1 * linkLen / velMax;
	totalT = totalT + deltaT;
	
	flowW = 0.0;
	flowTL = 0.0;
	for(i=0; i<totalLinks; i++) {
		temp = bubbles_Move(i);
		volumeW = volumeW + temp;
		flowW += linkSatu[i]*linkQ[i];
		flowTL += linkQ[i];
		bubbles_Kill(i);
	}
	for(i=0; i<totalLinks; i++) {
		bubbles_Create(i);
		bubbles_Kill(i);
		while (numberOfBubbles[i] > maxBubbles) bubbles_Merge_CM(i);
	}
	volumeW = 0.0;
	fluxTL = 0.0;
	fluxW = 0.0;
	for(i=0; i<totalNodes; i++) {
		fluxTL = fluxTL + nodeFluxTL[i];
		fluxW = fluxW + nodeFluxW[i];
		nodeFluxTL[i] = 0.0;
		nodeFluxW[i] = 0.0;
	}
}

/*Postscript Header*/
void pic_Post(FILE *fp, int sx, int sy){
	fprintf(fp,"%%!PS-Adobe-2.0\n");
	fprintf(fp,"%%Title: cluster\n");
	fprintf(fp,"%%%%BoundingBox: %d %d %d %d\n",-sx, -sy, 2*(xLen+1)*sx, (yLen+1)*sy);
	fprintf(fp,"%%Pages: 1\n");
	fprintf(fp,"%%EndComments\n");
	fprintf(fp,"%%Page: 1 1\n");
	fprintf(fp," .2 setflat\n");
	fprintf(fp," /P1{newpath 0.7 0 360 arc 0.0 setgray fill}def\n");
	fprintf(fp," /L0{newpath moveto lineto 0.9 setgray stroke}def\n");
	fprintf(fp," /L1{newpath moveto lineto 0.0 setgray stroke}def\n");
	fprintf(fp," /L2{newpath moveto lineto 0.9 0.0 0.0 setrgbcolor stroke}def\n");
	fprintf(fp," /L3{newpath moveto lineto 0.0 0.9 0.1 setrgbcolor stroke}def\n");
	fprintf(fp," /LG{newpath moveto lineto setgray stroke}def\n");      
	fprintf(fp," /LC{newpath moveto lineto setrgbcolor stroke}def\n"); 
	fprintf(fp," %%===============================================\n");
	fprintf(fp," %d %d translate\n", 0, sy);
	fprintf(fp," %d %d scale 0.5 setlinewidth\n", sx, sy);
	fprintf(fp," %%===============================================\n");
}

// Postscript header, for saturation picture
void ps_Header(FILE *fp, int xLen, int yLen, int sx, int sy){
	fprintf(fp,"%%!PS-Adobe-2.0\n");
	fprintf(fp,"%%Title: cluster\n");
	fprintf(fp,"%%%%BoundingBox: %d %d %d %d\n", 0, -2*sy, 2*(xLen+1)*sx, yLen*sy);
	fprintf(fp,"%%Pages: 1\n");
	fprintf(fp,"%%EndComments\n");
	fprintf(fp,"%%Page: 1 1\n");
	fprintf(fp," .2 setflat\n");
	fprintf(fp," /P1{newpath 0.7 0 360 arc 0.0 setgray fill}def\n");
	fprintf(fp," /L0{newpath moveto lineto 0.8 setgray stroke}def\n");
	fprintf(fp," /L1{newpath moveto lineto 0.0 setgray stroke}def\n");
	fprintf(fp," /L2{newpath moveto lineto 0.8 0.0 0.0 setrgbcolor stroke}def\n");
	fprintf(fp," /L3{newpath moveto lineto 0.0 0.9 0.1 setrgbcolor stroke}def\n");
	fprintf(fp," /B1{newpath moveto 1 0 rlineto 0 1 rlineto -1 0 rlineto closepath setgray fill}def\n");
	fprintf(fp," %%===============================================\n");
	fprintf(fp," %d %d translate\n", sx, 0);
	fprintf(fp," %d %d scale 0.6 setlinewidth\n", sx, sy);
	fprintf(fp," %%===============================================\n");
}

////Writing bubble distribution picture, showing in which links the wettability has been changed, and how much////
void draw_linkangles(char dir[80]) {
	int i, link, xt, j;
	double x1, y1, x2, y2, str, end, cl;
	char fn1[80];
	FILE *fp1;
	sprintf(fn1,"%s/wetting_angle:%d.eps", dir,timeStep, timeStep);

	fp1 = fopen(fn1,"w");
	pic_Post(fp1, 8, 8);

	for (link=0; link<totalLinks; link++) {
	
		xt = link%(2*xLen);
		x1 = xt + xt%2;
		x2 = 1 + ((link/2) % xLen) * 2;
		y1 = (link/2) / xLen - 1;
		y2 = y1 + 1;
		if (((link/2)/xLen)%2 == 1) {
			x1++;
			x2++;
		}

	cl=1-(CosTheta[link]+1)/2;
		fprintf(fp1,"%f %f %f %f %f LG\n", cl, x1, y1, x2, y2);
																				
}
	fprintf(fp1,"showpage\n");
	fclose(fp1);
}	

//// make picture of oil/water saturation in the network///////
void draw_LinkSat(char dir[80]) {
	int i,j,k;
	double sat;
	char fn1[80];
	FILE *fp1;
	sprintf(fn1,"%s/Linksaturation:%d.eps", dir,timeStep, timeStep);
	fp1 = fopen(fn1,"w");
	ps_Header(fp1, xLen, yLen , 8, 8);
	
	for (k=0; k<totalLinks; k++) {
	
		i=k%(2*xLen);
		j=k/(2*xLen);
		if(j%2 ==1) i++;
		if ((k+1)%(4*xLen) == 0) i -= 2*xLen;		

		sat=1-linkSatu[k];
		fprintf(fp1,"%.3f %d %d B1\n", sat, i, j-1);											
}
	fprintf(fp1,"showpage\n");
	fclose(fp1);
}	

////Writing flow distribution picture, showing flowrate in each link, on log scale, WITH link lattice in picture////
void draw_Flowdistr(char dir[80], double qmax) {
	int i, link, xt, j;
	double x1, y1, x2, y2, str, end;
	char fn1[80];
	FILE *fp1, *fp7, *fp8;
	sprintf(fn1,"%s/before.eps: %d ", dir,is, is);
	fp1 = fopen(fn1,"w");
	pic_Post(fp1, 8, 8);
	sprintf(fn1,"%s/after.eps: %d ", dir,is, is);
	fp7 = fopen(fn1,"w");
	pic_Post(fp7, 8, 8);
	sprintf(fn1,"%s/final.eps: %d ", dir,is, is);
	fp8 = fopen(fn1,"w");
	pic_Post(fp8, 8, 8);
	
	for (link=0; link<totalLinks; link++) {
	
		xt = link%(2*xLen);
		x1 = xt + xt%2;
		x2 = 1 + ((link/2) % xLen) * 2;
		y1 = (link/2) / xLen - 1;
		y2 = y1 + 1;
		if (((link/2)/xLen)%2 == 1) {
			x1++;
			x2++;
		}
		fprintf(fp1,"%f %f %f %f L0\n", x1, y1, x2, y2);
		fprintf(fp7,"%f %f %f %f L0\n", x1, y1, x2, y2);
		fprintf(fp8,"%f %f %f %f L0\n", x1, y1, x2, y2);
		
		//printout of flowrates BEFORE wettability alterations
	if(q_beforechange[link]/qmax > 0.1){                                 	                                                       		
			fprintf(fp1,"%f %f %f %f %f LG\n", 0.0, x1, y1, x2, y2);			 
			}
	      else if(q_beforechange[link]/qmax > 0.01 && q_beforechange[link]/qmax < 0.1){
		fprintf(fp1,"%f %f %f %f %f LG\n", 0.33, x1, y1, x2, y2); }                    												
		else if(q_beforechange[link]/qmax > 0.001 && q_beforechange[link]/qmax < 0.01){
		fprintf(fp1,"%f %f %f %f %f LG\n", 0.66, x1, y1, x2, y2); }


		//printout of flowrates directly after wettability alterations
	if(q_afterchange[link]/qmax > 0.1){                                   
          fprintf(fp7,"%f %f %f %f %f LG\n", 0.0, x1, y1, x2, y2);			 
	  }
	      else if(q_afterchange[link]/qmax > 0.01 && q_afterchange[link]/qmax < 0.1){
			fprintf(fp7,"%f %f %f %f %f LG\n", 0.33, x1, y1, x2, y2); }                      														
		else if(q_afterchange[link]/qmax > 0.001 && q_afterchange[link]/qmax < 0.01){
			fprintf(fp7,"%f %f %f %f %f LG\n", 0.66, x1, y1, x2, y2); }

	
		//printout of flowrates at steady state after alterations
	if(q_final[link]/qmax > 0.1){                  	                                                     					
			fprintf(fp8,"%f %f %f %f %f LG\n", 0.0, x1, y1, x2, y2);			 
			}
	      else if(q_final[link]/qmax > 0.01 && q_final[link]/qmax < 0.1){
		fprintf(fp8,"%f %f %f %f %f LG\n", 0.33, x1, y1, x2, y2); }                      														
		else if(q_final[link]/qmax > 0.001 && q_final[link]/qmax < 0.01){
		fprintf(fp8,"%f %f %f %f %f LG\n", 0.66, x1, y1, x2, y2); }
		}

	fprintf(fp1,"showpage\n");
	fclose(fp1);
	fprintf(fp7,"showpage\n");
	fclose(fp7);
	fprintf(fp8,"showpage\n");
	fclose(fp8);

}

////Writing flow distribution picture, showing flowrate in each link, on log scale, WITHOUT link lattice in picture////
void draw_Flowdistr2(char dir[80], double qmax) {
	int i, link, xt, j, k;
	double x1, y1, x2, y2, str, end;
	char fn1[80];
	FILE *fp1, *fp7, *fp8;
	sprintf(fn1,"%s/before.eps: %d ", dir,is, is);
	fp1 = fopen(fn1,"w");
	ps_Header(fp1, xLen, yLen , 8, 8);
	sprintf(fn1,"%s/after.eps: %d ", dir,is, is);
	fp7 = fopen(fn1,"w");
	ps_Header(fp7, xLen, yLen , 8, 8);
	sprintf(fn1,"%s/final.eps: %d ", dir,is, is);
	fp8 = fopen(fn1,"w");
	ps_Header(fp8, xLen, yLen , 8, 8);
	
	for (k=0; k<totalLinks; k++) {
	
		i=k%(2*xLen);
		j=k/(2*xLen);
		if(j%2 ==1) i++;
		if ((k+1)%(4*xLen) == 0) i -= 2*xLen;		
											
	
		//printout of flowrates BEFORE wettability alterations
		if(q_beforechange[k]/qmax > 0.1){
		fprintf(fp1,"%.3f %d %d B1\n", 0.0, i, j-1);			 
			}
	      else if(q_beforechange[k]/qmax > 0.01 && q_beforechange[k]/qmax < 0.1){
		fprintf(fp1,"%.3f %d %d B1\n", 0.33, i, j-1); 
		}                  												
		else if(q_beforechange[k]/qmax > 0.001 && q_beforechange[k]/qmax <0.01){
		fprintf(fp1,"%.3f %d %d B1\n", 0.66, i, j-1);
		}

		else if(q_beforechange[k]/qmax < 0.001){
		fprintf(fp1,"%.3f %d %d B1\n", 1.0, i, j-1); }

		//printout of flowrates directly after wettability alterations
		if(q_afterchange[k]/qmax > 0.1){                                   
       	 	  fprintf(fp7,"%.3f %d %d B1\n", 0.0, i, j-1);			 
		  }
	      else if(q_afterchange[k]/qmax > 0.01 && q_afterchange[k]/qmax < 0.1){
			fprintf(fp7,"%.3f %d %d B1\n", 0.33, i, j-1); }                      														
		else if(q_afterchange[k]/qmax > 0.001 && q_afterchange[k]/qmax < 0.01){
			fprintf(fp7,"%.3f %d %d B1\n", 0.66, i, j-1); }

		else if(q_afterchange[k]/qmax < 0.001){
		fprintf(fp7,"%.3f %d %d B1\n", 1.0, i, j-1); }

		//printout of flowrates at steady state after alterations
		if(q_final[k]/qmax > 0.1){
		fprintf(fp8,"%.3f %d %d B1\n", 0.0, i, j-1);			 
			}
	    	  else if(q_final[k]/qmax > 0.01 && q_final[k]/qmax < 0.1){
		fprintf(fp8,"%.3f %d %d B1\n", 0.33, i, j-1); }                      														
		else if(q_final[k]/qmax > 0.001 && q_final[k]/qmax < 0.01){
		fprintf(fp8,"%.3f %d %d B1\n", 0.66, i, j-1); }
		
		else if(q_final[k]/qmax < 0.001){
		fprintf(fp8,"%.3f %d %d B1\n", 1.0, i, j-1); }
}

	fprintf(fp1,"showpage\n");
	fclose(fp1);
	fprintf(fp7,"showpage\n");
	fclose(fp7);
	fprintf(fp8,"showpage\n");
	fclose(fp8);

}

/*Saving the bubble distribution only, not picture*/
void save_bblDist(char dirA[80]) {
	int i, link;
	double str, end, len;
	char fn1[80];
	FILE *fp1;
	sprintf(fn1,"%s/bdst_T-%08d.d", dirA, timeStep);
	fp1 = fopen(fn1,"w");
	for (link=0; link<totalLinks; link++) {
		fprintf(fp1,"%d\t%2d", link, numberOfBubbles[link]);
		for (i=0; i<numberOfBubbles[link]; i++) fprintf(fp1,"\t%e\t%e", bubbleStr[link][i], bubbleEnd[link][i]);
		fprintf(fp1,"\n");
	}
	fclose(fp1);
}

/*Saving local flow distribution*/
void save_FlowDist(char dirA[80]) {
      int link, xt, sl;
      double fl, maxFl, x1, y1, x2, y2, cl;
      char fn1[80];
      FILE *fp1;
      sprintf(fn1,"%s/fdist_T-%08d.d", dirA, timeStep);
      fp1 = fopen(fn1,"w");
      for (link=0; link<totalLinks; link++) fprintf(fp1,"%d\t%e\n", link, linkQ[link]);
      fclose(fp1);
}

/*Saving parameters*/
void save_Parameters(char dir[80]) {
	int i;
	char fn1[80], fn2[80];
	FILE *fp1, *fp2;
	sprintf(fn1,"%s/parameters.d", dir);
	sprintf(fn2,"%s/linkRadius.d", dir);
	fp1 = fopen(fn1,"w");
	fp2 = fopen(fn2,"w");
	fprintf(fp1,"%4d %4d %.2f %.2f %.2f %.3f %2d %.1e %.1e %.2e %.2f %.1f %.1f %d %d %e %e %2d %d\n\n",
		   xLen, yLen, linkLen, linkRadMin, linkRadMax, saturationW, maxBubbles, muWET, muNON, 
		   qRate, sTens, globalPressureA, globalPressureB, totalStepsA, totalStepsB, flipT, totalT, coord, dStep);
	fprintf(fp1,"System size: %d x %d\n\n", xLen, yLen);
	fprintf(fp1,"Link (mm):\n Length: %.1f\n Radius min: %.2f\n Radius max: %.2f\n\n", linkLen, linkRadMin, linkRadMax);
	fprintf(fp1,"Total Samples: %d\n\n", Samples);
	fprintf(fp1,"Initial W Saturation: %.3f\n\nMaximum Bubbles: %2d\n\n", saturationW, maxBubbles);
	fprintf(fp1,"Viscosities (100 Poise):\n mu_wt = %.3f\n mu_nw = %.3f\n\n", muWET, muNON);
	fprintf(fp1,"Flow rate (mm^3/s): %.2e\n\n", qRate);
	fprintf(fp1,"Surface Tension (dyne/mm): %.2f\n\n", sTens);
	fprintf(fp1,"Global pressures (dyne/mm^2 = 10 Pa):\n P_a = %.1f\n P_b = %.1f\n\n", globalPressureA, globalPressureB);
	fprintf(fp1,"Times:\n Total stepsA: %d.   Total stepsB: %d.   dStep: %d\n", totalStepsA, totalStepsB, dStep);
	fprintf(fp1," Total time : %e.  Flip time : %e\n", totalT, flipT);
	//Link Radius
	for (i=0; i<totalLinks; i++) fprintf(fp2,"%d\t%e\n", i, linkRad[i]);
	fclose(fp1); fclose(fp2);
}
